#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>

#include <linux/init.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/interrupt.h>

#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/sibyte/swarm.h>
#include <asm/sibyte/sb1250.h>
#include <asm/sibyte/sb1250_regs.h>
#include <asm/sibyte/sb1250_uart.h>
#include <asm/sibyte/sb1250_int.h>
#include <asm/sibyte/sb1250_dma.h>
#include <asm/sibyte/64bit.h>

#include <asm/system.h>   /* cli(), *_flags */
#include <asm/segment.h>  /* memcpy and such */
#include <asm/pgtable.h>

#include "bcm-internal.h"

#define DEVNAME "BCM"

#define R_MBOX			0xb00200c0
#define R_MBOX_CLEAR	0xb00200d0

#define PCI_MAP			0xf800000000ULL
#define DM				0xb0020b40

struct bcm_dev *bdev;

int verbose=1;
MODULE_PARM(verbose, "i");

u64 *dmdesc;
u64 dmphys;

static void rx_task(void *data){
	struct bcm_dev *bdev;
	struct bcm_transfer *rx;
	
	bdev=(struct bcm_dev *)data;
	rx=bdev->rx_finished;
	bdev->rx_finished=NULL;
	bdev->rx_finished_last=NULL;
	while(rx!=NULL){
		struct protocol *p;
		struct bcm_transfer *tr;
		p=find_protocol(bdev, rx->proto);
		
		if(p) {
			tr=kmalloc(sizeof(struct bcm_transfer), GFP_KERNEL);
			memcpy(tr, rx, sizeof(struct bcm_transfer));
			tr->dev=bdev->info;
			p->payload->rx_action(bdev->info, tr);
		} else {
			kfree(rx->buffer);
		}
		
		tr=rx;
		rx=rx->next;
		kfree(tr);
	}
	bdev->common->msg=__cpu_to_le32(3);
	pciconfig[0x24]=1;
}
		

static void rx_dm_handler(int irq, void *dev_id, struct pt_regs *regs) {
	int tail, head;
	u64 status;
	struct bcm_dev *bdev;
	
	bdev=(struct bcm_dev *)dev_id;
	status=in64(DM);
	if(status==0x1000000000000000ULL)return;
	if(verbose>2)printk(KERN_ERR DEVNAME ": rx_dm_handler, status=%016Lx\n",status);
	bdev->rx_dma_running=0;
	out64(0, DM);
	if(status&(1ULL<<61)) {
		out64(1ULL<<62, DM);	/* reset DM */
	} else {
		struct bcm_transfer *tr;
		tr=kmalloc(sizeof(struct bcm_transfer), GFP_ATOMIC);
		memcpy(tr, &bdev->rx_in_progress, sizeof(struct bcm_transfer));
		tr->next=NULL;
		if(bdev->rx_finished_last==NULL) {
			bdev->rx_finished=tr;
		} else {
			bdev->rx_finished_last->next=tr;
		}
		bdev->rx_finished_last=tr;
		schedule_task(&bdev->tq_rx);
	}
			
	tail=__le32_to_cpu(bdev->common->tx_tail);
	tail++;
	if(tail==bdev->tx_size)tail=0;
	bdev->common->tx_tail=__le32_to_cpu(tail);
	head=__le32_to_cpu(bdev->common->tx_head);
		
	if(verbose>2)printk(KERN_ERR DEVNAME": tail=%i  head=%i\n",tail, head);
	if(head != tail) { /* queue not empty yet */
		int size;
		u64 src, dest;
		
		size=__le32_to_cpu(bdev->tx_ring[tail].len);
		src=__le64_to_cpu(bdev->tx_ring[tail].phys) | PCI_MAP;
		
		bdev->rx_in_progress.buffer=kmalloc(size, GFP_ATOMIC);
		bdev->rx_in_progress.len=size;
		bdev->rx_in_progress.proto=__le32_to_cpu(bdev->tx_ring[tail].proto);
		
		dest=virt_to_phys(bdev->rx_in_progress.buffer);
		dmdesc[0]=dest | (0x0406ULL<<40); /* dest cachable, interrupt */
		dmdesc[1]=src | ((u64)size << 40);
		out64(dmphys | (1ULL<<40) | (1ULL<<61) | (1ULL<<63), DM);
		out64(1, DM+8);
		bdev->rx_dma_running=1;
	}
}

void mbox_rx_task(unsigned long data) {
	int head, tail;
	struct bcm_dev *bdev;

	if(verbose>1)printk(KERN_DEBUG DEVNAME ": data tx interrupt\n");
	bdev=*(struct bcm_dev **)data;
	head=__le32_to_cpu(bdev->common->tx_head);
	tail=__le32_to_cpu(bdev->common->tx_tail);
	if(head==tail) {
		if(verbose>1)
			printk(KERN_WARNING DEVNAME ": data tx interrupt while tx queue is empty.\n");
	} else if(!bdev->rx_dma_running){
		int size;
		u64 src, dest;
		
		size=__le32_to_cpu(bdev->tx_ring[tail].len);
		src=__le64_to_cpu(bdev->tx_ring[tail].phys) | PCI_MAP;
		bdev->rx_in_progress.buffer=kmalloc(size, GFP_ATOMIC);
		bdev->rx_in_progress.len=size;
		bdev->rx_in_progress.flags=__le32_to_cpu(bdev->tx_ring[tail].flags);
		bdev->rx_in_progress.proto=__le32_to_cpu(bdev->tx_ring[tail].proto);
		
		dest=virt_to_phys(bdev->rx_in_progress.buffer);
		dmdesc[0]=dest | (0x0406ULL<<40); /* dest cachable, interrupt */
		dmdesc[1]=src | ((u64)size << 40);
		out64(dmphys | (1ULL<<40) | (1ULL<<61) | (1ULL<<63), DM);
		out64(1, DM+8);
		bdev->rx_dma_running=1;
		if(verbose>2)printk(KERN_ERR DEVNAME ": status is %016Lx   %016Lx   %016Lx\n",
				in64(DM+0x18),dmdesc[0],dmdesc[1]);
	}
}

DECLARE_TASKLET(mbox_rx_tasklet, mbox_rx_task, (unsigned long)&bdev);

static void callback(int irq) {

	switch(irq) {
		case 2:
			if(ldt_master) {
				printk(KERN_WARNING DEVNAME ": Master card received init interrupt.\n");
			} else {
			}
			break;
		case 4:
			break;
		case 8:
			break;
		
	}
	while(mbox) {
	
		if(mbox&INT_ACK) {
			bdev->common->msg=0;
			pciconfig[0x24]=0;
			out64(INT_ACK, R_MBOX_CLEAR);
			if(verbose>2)printk(KERN_DEBUG DEVNAME ": acknowledge interrupt\n");
		}
		if(mbox&INT_INIT) {
			bdev->tx_size=__le32_to_cpu(bdev->common->tx_size);
			bdev->rx_size=__le32_to_cpu(bdev->common->rx_size);
			bdev->tx_ring=ioremap(__le64_to_cpu(bdev->common->tx_phys)+ PCI_MAP, 
					bdev->tx_size*sizeof(ring_element));
			bdev->rx_ring=ioremap(__le64_to_cpu(bdev->common->rx_phys)+ PCI_MAP, 
					bdev->tx_size*sizeof(ring_element));
			if(verbose>2)printk(KERN_DEBUG DEVNAME ": init interrupt  rx=%08Lx:%i  tx=%08Lx:%i\n",
					__le64_to_cpu(bdev->common->rx_phys), bdev->rx_size, 
					__le64_to_cpu(bdev->common->tx_phys), bdev->tx_size);
			out64(INT_INIT, R_MBOX_CLEAR);
			bdev->common->msg=__cpu_to_le32(1);
			pciconfig[0x24]=1; /* all ready */

		}
		if(mbox&INT_TX) {
			out64(INT_TX, R_MBOX_CLEAR);
			tasklet_schedule(&mbox_rx_tasklet);
		}
		if(mbox&INT_RX) {
			out64(INT_RX, R_MBOX_CLEAR);
			tasklet_schedule(&mbox_tx_tasklet);
		}

		if(mbox&INT_UNK) {
			if(verbose>2)printk(KERN_WARNING DEVNAME 
					": unexpected interrupt, mbox[31:16]=%04Lx\n", mbox>>16);
			out64(INT_UNK, R_MBOX_CLEAR);
		}
		mbox=in64(R_MBOX) & 0xffff0000ULL;
	}
		
}

static int __init s(void) {
	unsigned long addr;
	unsigned long p;

	bdev=kmalloc(sizeof(struct bcm_dev), GFP_KERNEL);
	
	dmdesc=kmalloc(32, GFP_KERNEL);
	dmphys=virt_to_bus(dmdesc);
	bdev->tx_dma_running=0;
	bdev->rx_dma_running=0;

	bdev->common = (struct common_data *) ldt_buf;

	p=virt_to_phys(bdev->common);

	if(lst_master) {
		bdev->common->signature[0]='B';
		bdev->common->signature[1]='C';
		bdev->common->signature[2]='M';
		bdev->common->signature[3]=174;
	}

	bdev->tx_size=-1;
	bdev->rx_size=-1;
	bdev->tq_rx.data=bdev;
	bdev->tq_rx.routine=rx_task;
	bdev->tq_rx.sync=0;

	bdev->tq_tx.data=bdev;
	bdev->tq_tx.routine=tx_task;
	bdev->tq_tx.sync=0;
	bdev->tx_card=NULL;
	bdev->rx_card=NULL;
	bdev->tx_finished=NULL;
	bdev->tx_finished_last=NULL;
	bdev->rx_finished=NULL;
	bdev->rx_finished_last=NULL;
		
	bdev->prot=NULL;

	request_irq(K_INT_DM_CH_2, rx_dm_handler, SA_SHIRQ, DEVNAME, bdev);
	ldt_callback=callback;

	if(ldt_master) {
		ldt_interrupt(2<<13);
	}
	
	return 0;
}

static void __exit e(void) {
	if(bdev->common) {
		free_pages((unsigned long)bdev->common, 8);
		if(pciconfig) {
			pciconfig[0x11]=0;
			pciconfig[0x24]=0;
		}
	}
	free_irq(K_INT_MBOX_2, bdev);
	free_irq(K_INT_DM_CH_0, bdev);
	free_irq(K_INT_DM_CH_1, bdev);
	kfree(dmdesc);
	kfree(bdev);
	return ;
}
module_init(s);
module_exit(e);


/*
 * external interface
 */
struct dev_info di;

int bcm_register_payload(struct dev_info *dev, struct bcm_payload *pl) {
	struct protocol *p;
	struct bcm_dev * bdev=dev->bcm;
	
	p=find_protocol(bdev, pl->proto);
	if(p!=NULL) {
		p->payload=pl;
	} else {
		p=bdev->prot;
		bdev->prot=kmalloc(sizeof(struct protocol), GFP_KERNEL);
		bdev->prot->next=p;
		bdev->prot->payload=pl;
		bdev->prot->proto=pl->proto;
	}
	return 0;
}

int bcm_unregister_payload(struct dev_info *dev, struct bcm_payload *pl) {
	struct protocol *p;
	struct bcm_dev *bdev=dev->bcm;
	p=bdev->prot;
	while(p!=NULL && p->payload!=pl)p=p->next;
	if(p!=NULL) {
		if(p==bdev->prot) {
			bdev->prot=p->next;
			kfree(p);
		} else {
			struct protocol *t;
			t=bdev->prot;
			while(t!=NULL && t->next!=p)t=t->next;
			if(t->next==p) {
				t->next=p->next;
				kfree(p);
			}
		}
		return 0;
	}
	return -1;
}

int bcm_transfer_data(struct bcm_transfer *tr) {
	struct bcm_dev *bdev;
	struct bcm_transfer *t, *s;

	bdev=tr->dev->bcm;
	s=kmalloc(sizeof(struct bcm_transfer), GFP_KERNEL);
	memcpy(s, tr, sizeof(struct bcm_transfer));
	s->next=NULL;
	
	disable_irq(K_INT_DM_CH_1);
	t=bdev->tx_card;
	if(t==NULL) {
		bdev->tx_card=s;
	} else {
		while(t->next!=NULL){
			t=t->next;
		}
		t->next=s;
	}
	enable_irq(K_INT_DM_CH_1);
	if(!bdev->tx_dma_running)start_tx(bdev);
	return 0;
}

EXPORT_SYMBOL(bcm_register_payload);
EXPORT_SYMBOL(bcm_unregister_payload);
EXPORT_SYMBOL(bcm_transfer_data);
